@@ -60,6 +60,7 @@ u16 checksum( byte *p, unsigned n );
 u16 checksum_mpi( MPI a );
 u16 checksum_mpi_counted_nbits( MPI a );
 u32 buffer_to_u32( const byte *buffer );
+const byte *get_session_marker( size_t *rlen );
 
 /*-- helptext.c --*/
 void display_online_help( const char *keyword );
@@ -153,6 +154,9 @@ int decrypt_message( const char *filename );
 int hash_datafiles( MD_HANDLE md, MD_HANDLE md2,
 		    STRLIST files, const char *sigfilename, int textmode );
 
+/*-- pipemode.c --*/
+void run_in_pipemode (void);
+
 /*-- signal.c --*/
 void init_signals(void);
 void pause_on_sigusr( int which );
