@@ -1,5 +1,5 @@
 /* ringedit.c -  Function for key ring editing
- *	Copyright (C) 1998 Free Software Foundation, Inc.
+ *	Copyright (C) 1998, 2000 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -63,10 +63,6 @@
 #include "i18n.h"
 
 
-#ifdef MKDIR_TAKES_ONE_ARG
-# undef mkdir
-# define mkdir(a,b) mkdir(a)
-#endif
 
 
 struct resource_table_struct {
@@ -185,7 +181,6 @@ enum_keyblock_resources( int *sequence, int secret )
 }
 
 
-
 /****************
  * Register a resource (which currently may only be a keyring file).
  * The first keyring which is added by this function is
@@ -291,24 +286,13 @@ add_keyblock_resource( const char *url, int force, int secret )
 	    *last_slash_in_filename = 0;
 
 	    if( access(filename, F_OK) ) {
-		if( strlen(filename) >= 7
-		    && !strcmp(filename+strlen(filename)-7, "/.gnupg") ) {
-		    if( mkdir(filename, S_IRUSR|S_IWUSR|S_IXUSR) )
-		    {
-			log_error( _("%s: can't create directory: %s\n"),
-				  filename, strerror(errno));
-			rc = G10ERR_OPEN_FILE;
-			goto leave;
-		    }
-		    else if( !opt.quiet )
-			log_info( _("%s: directory created\n"), filename );
-		    copy_options_file( filename );
-		}
-		else
-		{
-		    rc = G10ERR_OPEN_FILE;
-		    goto leave;
-		}
+		/* on the first time we try to create the default homedir and
+		 * in this case the process will be terminated, so that on the
+		 * next invocation it can read the options file in on startup
+		 */
+		try_make_homedir( filename );
+		rc = G10ERR_OPEN_FILE;
+		goto leave;
 	    }
 
 	    *last_slash_in_filename = '/';
@@ -435,6 +419,35 @@ get_keyblock_handle( const char *filename, int secret, KBPOS *kbpos )
 }
 
 
+/****************
+ * Return the filename of the firstkeyblock resource which is intended
+ * for write access. This will either be the default resource or in
+ * case this is not writable one of the others.  If no writable is found,
+ * the default filename in the homedirectory will be returned.
+ * Caller must free, will never return NULL.
+ */
+char *
+get_writable_keyblock_file( int secret )
+{
+    int i = secret? default_secret_resource : default_public_resource;
+
+    if( resource_table[i].used && !resource_table[i].secret == !secret ) {
+	if( !access( resource_table[i].fname, R_OK|W_OK ) ) {
+	    return m_strdup( resource_table[i].fname );
+	}
+    }
+    for(i=0; i < MAX_RESOURCES; i++ ) {
+	if( resource_table[i].used && !resource_table[i].secret == !secret ) {
+	    if( !access( resource_table[i].fname, R_OK|W_OK ) ) {
+		return m_strdup( resource_table[i].fname );
+	    }
+	}
+    }
+    /* Assume the home dir is always writable */
+    return  make_filename(opt.homedir, secret? "secring.gpg"
+					     : "pubring.gpg", NULL );
+}
+
 
 /****************
  * Search a keyblock which starts with the given packet and puts all
@@ -737,7 +750,9 @@ enum_keyblocks( int mode, KBPOS *kbpos, KBNODE *ret_root )
 
     if( !mode || mode == 5 || mode == 100 ) {
 	int i;
+
 	kbpos->fp = NULL;
+	kbpos->rt = rt_UNKNOWN;
 	if( !mode ) {
 	    kbpos->secret = 0;
 	    i = 0;
@@ -1055,7 +1070,13 @@ cmp_seckey( PKT_secret_key *req_sk, PKT_secret_key *sk )
 
     n = pubkey_get_nskey( req_sk->pubkey_algo );
     for(i=0; i < n; i++ ) {
-	if( mpi_cmp( req_sk->skey[i], sk->skey[i] ) )
+        /* Note: becuase v4 protected keys have nothing in the
+         * mpis except for the first one, we skip all NULL MPIs.
+         * This might not be always correct in cases where the both
+         * keys do not match in their secret parts but we can ignore that
+         * because the need for this function is quite ugly. */
+	if( req_sk->skey[1] && sk->skey[i]
+             && mpi_cmp( req_sk->skey[i], sk->skey[i] ) )
 	    return -1;
     }
     return 0;
@@ -1085,7 +1106,7 @@ keyring_search( PACKET *req, KBPOS *kbpos, IOBUF iobuf, const char *fname )
     int rc;
     PACKET pkt;
     int save_mode;
-    ulong offset;
+    off_t offset;
     int pkttype = req->pkttype;
     PKT_public_key *req_pk = req->pkt.public_key;
     PKT_secret_key *req_sk = req->pkt.secret_key;
@@ -1167,9 +1188,9 @@ keyring_read( KBPOS *kbpos, KBNODE *ret_root )
     }
 
     if( !kbpos->valid )
-	log_debug("kbpos not valid in keyring_read, want %d\n", (int)kbpos->offset );
+       log_debug("kbpos not valid in keyring_read\n" );
     if( iobuf_seek( a, kbpos->offset ) ) {
-	log_error("can't seek to %lu\n", kbpos->offset);
+       log_error("can't seek\n");
 	iobuf_close(a);
 	return G10ERR_KEYRING_OPEN;
     }
